@@ -7,7 +7,7 @@
  *
  */
 
-/* Copyright (C) 2010, 2011 Jose E. Marchesi */
+/* Copyright (C) 2010, 2011, 2012 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include <xalloc.h>
 #include <gettext.h>
 #define _(str) gettext (str)
+#include <getpass.h>
 
 #include <rec.h>
 #include <recutl.h>
@@ -47,6 +48,7 @@ bool       recsel_print_values = false;
 bool       recsel_print_row    = false;
 char      *recutl_sex_str      = NULL;
 rec_sex_t  recutl_sex          = NULL;
+char      *recutl_quick_str    = NULL;
 char      *recsel_fex_str      = NULL;
 rec_fex_t  recsel_fex          = NULL;
 char      *recutl_type         = NULL;
@@ -54,10 +56,11 @@ bool       recsel_collapse     = false;
 bool       recsel_count        = false;
 bool       recutl_insensitive  = false;
 bool       recsel_descriptors  = false;
-size_t     recutl_num          = -1;
 rec_field_name_t recutl_sort_by_field = NULL;
 rec_writer_mode_t recsel_write_mode = REC_WRITER_NORMAL;
 char      *recsel_password     = NULL;
+bool       recsel_uniq         = false;
+size_t     recutl_random       = 0;
 
 /*
  * Command line options management.
@@ -75,7 +78,8 @@ enum
   DESCRIPTOR_ARG,
   PRINT_SEXPS_ARG,
   SORT_ARG,
-  PASSWORD_ARG
+  PASSWORD_ARG,
+  UNIQ_ARG
 };
 
 static const struct option GNU_longOptions[] =
@@ -91,6 +95,7 @@ static const struct option GNU_longOptions[] =
     {"print-sexps", no_argument, NULL, PRINT_SEXPS_ARG},
     {"sort", required_argument, NULL, SORT_ARG},
     {"password", required_argument, NULL, PASSWORD_ARG},
+    {"uniq", no_argument, NULL, UNIQ_ARG},
     {NULL, 0, NULL, 0}
   };
 
@@ -104,7 +109,7 @@ recutl_print_help (void)
   /* TRANSLATORS: --help output, recsel synopsis.
      no-wrap */
   printf (_("\
-Usage: recsel [OPTION]... [-t TYPE] [-n NUM | -e RECORD_EXPR] [-c | (-p|-P) FIELD_EXPR] [FILE]...\n"));
+Usage: recsel [OPTION]... [-t TYPE] [-n INDEXES | -e RECORD_EXPR | -q EXPR | -m NUM] [-c | (-p|-P) FIELD_EXPR] [FILE]...\n"));
 
   /* TRANSLATORS: --help output, recsel arguments.
      no-wrap */
@@ -118,7 +123,8 @@ Select and print rec data.\n"), stdout);
   -d, --include-descriptors           print record descriptors along with the matched\n\
                                         records.\n\
   -C, --collapse                      do not section the result in records with newlines.\n\
-  -S, --sort=FIELD                    sort the output by the specified field.\n"),
+  -S, --sort=FIELD                    sort the output by the specified field.\n\
+  -U, --uniq                          remove duplicated fields in the output records.\n"),
          stdout);
 
 #if defined REC_CRYPT_SUPPORT
@@ -170,7 +176,7 @@ recsel_parse_args (int argc,
   while ((ret = getopt_long (argc,
                              argv,
                              RECORD_SELECTION_SHORT_ARGS
-                             "S:Cdcp:P:R:s:",
+                             "S:Cdcp:P:R:s:U",
                              GNU_longOptions,
                              NULL)) != -1)
     {
@@ -190,6 +196,12 @@ recsel_parse_args (int argc,
             recsel_write_mode = REC_WRITER_SEXP;
             break;
           }
+        case UNIQ_ARG:
+        case 'U':
+          {
+            recsel_uniq = true;
+            break;
+          }
         case PASSWORD_ARG:
         case 's':
           {
@@ -281,10 +293,6 @@ recsel_parse_args (int argc,
 
         }
     }
-
-  
-  /* Set the sorting criteria for the parser.  */
-  recutl_sorting_parser (true, recutl_type, recutl_sort_by_field);
 }
 
 bool
@@ -298,19 +306,56 @@ recsel_process_data (rec_db_t db)
   rec_writer_t writer;
   bool parse_status;
   bool wrote_descriptor;
-  rec_rset_elem_t elem_rset;
+  rec_mset_iterator_t iter;
 
   ret = true;
 
   writer = rec_writer_new (stdout);
-  if (recsel_password)
+
+#if defined REC_CRYPT_SUPPORT
+
+  /* Set the password in the writer.  If recsel was called
+     interactively and with an empty -s, was not used then prompt the
+     user for it.  Otherwise use the password specified in the command
+     line if any.  */
+
+  if (!recsel_password
+      && (recutl_type || (rec_db_size (db) == 1))
+      && recutl_interactive ())
+    {
+      rec_rset_t rset;
+      rec_fex_t confidential_fields;
+
+      if (recutl_type)
+        {
+          rset = rec_db_get_rset_by_type (db, recutl_type);
+        }
+      else
+        {
+          rset = rec_db_get_rset (db, 0);
+        }
+
+      confidential_fields = rec_rset_confidential (rset);
+      if (rec_fex_size (confidential_fields) > 0)
+        {
+          recsel_password = getpass (_("Password: "));
+        }
+
+      rec_fex_destroy (confidential_fields);
+    }
+
+  /* Note that the password must be at least one character long.  */
+
+  if (recsel_password && (strlen (recsel_password) > 0))
     {
       rec_writer_set_password (writer,
                                recsel_password);
     }
 
+#endif /* REC_CRYPT_SUPPORT */
+
   /* If the database contains more than one type of records and the
-     user did'nt specify the recutl_type then ask the user to clear
+     user did'nt specify the recutl_type then ask the user to clarify
      the request.  */
   if (!recutl_type && (rec_db_size (db) > 1))
     {
@@ -354,26 +399,50 @@ recsel_process_data (rec_db_t db)
           continue;
         }
           
-      /*  Process this record set.  */
+      /* Process this record set.  */
+
+      /* If the user requested to print random records, calculate them
+         now for this record set.  */
+
+      if (recutl_random > 0)
+        {
+          recutl_reset_indexes ();
+          recutl_index_add_random (recutl_random, rec_rset_num_records (rset));
+        }
+
+      rec_rset_sort (rset, recutl_sort_by_field);
+
       num_rec = -1;
-      elem_rset = rec_rset_null_elem ();
-      while (rec_rset_elem_p (elem_rset = rec_rset_next_record (rset, elem_rset)))
+      iter = rec_mset_iterator (rec_rset_mset (rset));
+      while (rec_mset_iterator_next (&iter, MSET_RECORD, (const void **) &record, NULL))
         {
-          record = rec_rset_elem_record (elem_rset);
           num_rec++;
 
           /* Shall we skip this record?  */
-          if (((recutl_num != -1) && (recutl_num != num_rec))
-              || (recutl_sex_str && !(rec_sex_eval (recutl_sex, record, &parse_status)
-                                      && parse_status)))
+
+          if (recutl_quick_str)
             {
-              if (recutl_sex_str && (!parse_status))
+              if (!rec_record_contains_value (record,
+                                              recutl_quick_str,
+                                              recutl_insensitive))
                 {
-                  recutl_error (_("evaluating the selection expression.\n"));
-                  return false;
+                  continue;
                 }
+            }
+          else
+            {
+              if (((recutl_num_indexes() != 0) && (!recutl_index_p (num_rec)))
+                  || (recutl_sex_str && !(rec_sex_eval (recutl_sex, record, &parse_status)
+                                          && parse_status)))
+                {
+                  if (recutl_sex_str && (!parse_status))
+                    {
+                      recutl_error (_("evaluating the selection expression.\n"));
+                      return false;
+                    }
       
-              continue;
+                  continue;
+                }
             }
 
           /* Process this record.  */
@@ -386,13 +455,22 @@ recsel_process_data (rec_db_t db)
             {
               char *output = NULL;
 
+              /* Remove duplicated fields in the record if requested
+                 by the user.  */
+
+              if (recsel_uniq)
+                {
+                  rec_record_uniq (record);
+                }
+
               if (recsel_fex_str)
                 {
                   output = recutl_eval_field_expression (recsel_fex,
                                                          record,
                                                          recsel_write_mode,
                                                          recsel_print_values,
-                                                         recsel_print_row);
+                                                         recsel_print_row,
+                                                         recsel_password);
                 }
 
               /* Insert a newline?  */
@@ -432,6 +510,8 @@ recsel_process_data (rec_db_t db)
               written++;
             }
         }
+
+      rec_mset_iterator_free (&iter);
     }
 
   if (recsel_count)
